import { Button } from '@/lib';
import type { Meta, StoryObj } from '@storybook/react';
import { useState } from 'react';
import { Menu, MenuCheckboxItem, MenuContent, MenuItem, MenuLabel, MenuRadioGroup, MenuRadioItem, MenuSeparator, MenuTrigger } from './';

const meta = {
  title: 'Components/Menu',
  component: Menu,
  tags: ['autodocs'],
  parameters: {
    layout: 'centered',
    docs: {
      description: {
        component: `
A flexible menu component that supports nested submenus, checkboxes, radio groups, section labels, and custom triggers.
Features include:
- Click or hover activation
- Keyboard navigation
- Custom placement options
- Nested submenus
- Section labels for better grouping
- Checkbox and radio items
- Custom triggers
- Disabled states
        `,
      },
    },
  },
  argTypes: {
    trigger: {
      control: 'radio',
      options: ['click', 'hover'],
      description: 'The interaction method to open the menu',
      table: {
        defaultValue: { summary: 'click' },
      },
    },
    placement: {
      control: 'select',
      options: [
        'top',
        'top-start',
        'top-end',
        'bottom',
        'bottom-start',
        'bottom-end',
        'right',
        'right-start',
        'right-end',
        'left',
        'left-start',
        'left-end',
      ],
      description: 'The preferred placement of the menu',
      table: {
        defaultValue: { summary: 'bottom-start' },
      },
    },
    isOpen: {
      control: 'boolean',
      description: 'Controls the open state of the menu',
    },
    onOpenChange: {
      action: 'onOpenChange',
      description: 'Callback fired when the open state changes',
    },
  },
  args: {
    trigger: 'click',
    placement: 'bottom-start',
  },
} satisfies Meta<typeof Menu>;

export default meta;
type Story = StoryObj<typeof Menu>;

export const Basic: Story = {
  name: 'Basic Usage',
  render: args => (
    <Menu {...args}>
      <MenuTrigger label="Open Menu" />
      <MenuContent>
        <MenuLabel label="File Options" />
        <MenuItem label="New File" onClick={() => console.log('New File')} />
        <MenuItem label="Open..." onClick={() => console.log('Open')} />
        <MenuItem label="Save" onClick={() => console.log('Save')} />
        <MenuSeparator />
        <MenuItem label="Exit" onClick={() => console.log('Exit')} />
      </MenuContent>
    </Menu>
  ),
};

export const WithDisabledItems: Story = {
  name: 'With Disabled Items',
  render: args => (
    <Menu {...args}>
      <MenuTrigger label="Edit" />
      <MenuContent>
        <MenuLabel label="Editing" />
        <MenuItem label="Undo" onClick={() => console.log('Undo')} />
        <MenuItem label="Redo" disabled onClick={() => console.log('Redo')} />
        <MenuSeparator />
        <MenuItem label="Cut" disabled onClick={() => console.log('Cut')} />
        <MenuItem label="Copy" onClick={() => console.log('Copy')} />
        <MenuItem label="Paste" onClick={() => console.log('Paste')} />
      </MenuContent>
    </Menu>
  ),
};

export const WithCheckboxItems: Story = {
  name: 'With Checkbox Items',
  render: function CheckboxStory(args) {
    const [items, setItems] = useState({
      spellcheck: true,
      readonly: false,
      hidden: false,
      disabled: true,
    });

    return (
      <Menu {...args}>
        <MenuTrigger label="Options" />
        <MenuContent>
          <MenuLabel label="Preferences" />
          <MenuCheckboxItem
            checked={items.spellcheck}
            onCheckedChange={checked => setItems(prev => ({ ...prev, spellcheck: checked }))}
            label="Spellcheck"
          />

          <MenuCheckboxItem
            checked={items.readonly}
            onCheckedChange={checked => setItems(prev => ({ ...prev, readonly: checked }))}
            label="Read-only"
          />
          <MenuCheckboxItem
            checked={items.hidden}
            onCheckedChange={checked => setItems(prev => ({ ...prev, hidden: checked }))}
            label="Hidden"
          />
          <MenuCheckboxItem
            disabled
            checked={items.disabled}
            onCheckedChange={checked => setItems(prev => ({ ...prev, disabled: checked }))}
            label="Disabled"
          />
        </MenuContent>
      </Menu>
    );
  },
};

export const WithRadioItems: Story = {
  name: 'With Radio Items',
  render: function RadioStory(args) {
    const [size, setSize] = useState('medium');

    return (
      <Menu {...args}>
        <MenuTrigger label="Select Size" />
        <MenuContent>
          <MenuLabel label="Size Options" />
          <MenuRadioGroup value={size} onValueChange={setSize}>
            <MenuRadioItem value="small" label="Small" />
            <MenuRadioItem value="medium" label="Medium" />
            <MenuRadioItem value="large" label="Large" />
          </MenuRadioGroup>
        </MenuContent>
      </Menu>
    );
  },
};

export const NestedSubmenus: Story = {
  name: 'Nested Submenus',
  render: args => (
    <Menu {...args}>
      <MenuTrigger label="File" />
      <MenuContent>
        <MenuLabel label="Actions" />
        <MenuItem closeOnClick={true} iconRight="icon-[ic--outline-new-releases]" label="New" onClick={() => alert('New clicked')} />
        <MenuItem closeOnClick={true} iconRight="icon-[ic--outline-new-releases]" label="New 2" />
        <MenuItem closeOnClick={false} iconLeft="icon-[ic--baseline-content-copy]" rightSlot={<div>Ctrl + C</div>}>
          <div>Copy</div>
        </MenuItem>
        <Menu>
          <MenuTrigger label="Open Recent" />
          <MenuContent>
            <MenuItem label="Document 1.txt" />
            <MenuItem label="Document 2.txt" />
            <MenuSeparator />
            <MenuItem label="Document 1.txt" />
            <MenuItem label="Document 2.txt" />
            <MenuItem label="Document 1.txt" />
            <MenuItem label="Document 2.txt" />
            <Menu placement="right-start">
              <MenuTrigger label="More Options" />
              <MenuContent>
                <MenuItem label="Clear List" />
                <MenuItem label="Show Folder" />
                <MenuItem label="Show Folder1" />
                <MenuItem label="Show Folder2" />
                <MenuItem label="Show Folder3" />
                <MenuItem label="Show Folder4" />
                <MenuItem label="Show Folder5" />
                <MenuItem label="Show Folder6" onClick={() => alert('New clicked')} />
              </MenuContent>
            </Menu>
          </MenuContent>
        </Menu>
        <MenuSeparator />
        <MenuItem label="Save" onClick={() => console.log('Save')} />
        <MenuItem label="Save As..." onClick={() => console.log('Save As')} />
      </MenuContent>
    </Menu>
  ),
};

export const CustomTrigger: Story = {
  name: 'Custom Trigger',
  render: args => (
    <Menu {...args}>
      <MenuTrigger as="div">
        <Button size="xl" rounded variantType="primary" iconComponent="icon-[ic--baseline-more-vert]" />
      </MenuTrigger>
      <MenuContent>
        <MenuLabel label="Custom Actions" />
        <MenuItem label="Option 1" />
        <MenuItem label="Option 2" />
        <MenuItem label="Option 3" />
      </MenuContent>
    </Menu>
  ),
};

export const HoverActivation: Story = {
  name: 'Hover Activation',
  args: {
    trigger: 'hover',
  },
  render: args => (
    <Menu {...args}>
      <MenuTrigger label="Hover Me" />
      <MenuContent>
        <MenuLabel label="Quick Actions" />
        <MenuItem label="Option 1" />
        <MenuItem label="Option 2" />
        <Menu trigger="hover">
          <MenuTrigger label="More Options" />
          <MenuContent>
            <MenuItem label="Sub Option 1" />
            <MenuItem label="Sub Option 2" />
          </MenuContent>
        </Menu>
      </MenuContent>
    </Menu>
  ),
};

export const RightClickAtCursor: Story = {
  name: 'Right click at cursor',
  render: args => (
    <Menu atCursor {...args}>
      <MenuTrigger rightClick>
        <div className="flex justify-center items-center rounded bg-blue-300 w-52 h-52">Right click me</div>
      </MenuTrigger>
      <MenuContent>
        <MenuLabel label="Quick Actions" />
        <MenuItem label="Option 1" />
        <MenuItem label="Option 2" />
        <Menu trigger="hover">
          <MenuTrigger label="More Options" />
          <MenuContent>
            <MenuItem label="Sub Option 1" />
            <MenuItem label="Sub Option 2" />
          </MenuContent>
        </Menu>
      </MenuContent>
    </Menu>
  ),
};
