import { Icon } from '@/lib';
import type { OpenChangeReason, Placement } from '@floating-ui/react';
import {
  arrow,
  autoUpdate,
  flip,
  FloatingFocusManager,
  FloatingList,
  FloatingPortal,
  FloatingTree,
  offset,
  safePolygon,
  shift,
  useClick,
  useClientPoint,
  useDismiss,
  useFloating,
  useFloatingNodeId,
  useFloatingParentNodeId,
  useFloatingTree,
  useHover,
  useInteractions,
  useListItem,
  useListNavigation,
  useMergeRefs,
  useRole,
  useTypeahead,
} from '@floating-ui/react';
import * as React from 'react';
import { useControllableState } from '../primitives';
import styles from './menu.module.scss';

// --- Menu Context and Provider ---
export interface MenuContextProps {
  activeIndex: number | null;
  setActiveIndex: React.Dispatch<React.SetStateAction<number | null>>;
  getItemProps: (userProps?: React.HTMLProps<HTMLElement>) => Record<string, unknown>;
  isOpen: boolean;
  setIsOpen: (value: boolean, event?: Event, reason?: OpenChangeReason) => void;
  trigger: 'rightClick' | 'click' | 'hover';
  floating: {
    x: number | null;
    y: number | null;
    strategy: 'absolute' | 'fixed';
    refs: ReturnType<typeof useFloating>['refs'];
    context: ReturnType<typeof useFloating>['context'];
    interactions: ReturnType<typeof useInteractions>;
  };
  elementsRef: React.RefObject<Array<HTMLButtonElement | null>>;
  labelsRef: React.RefObject<Array<string | null>>;
  isNested: boolean;
  parent: MenuContextProps | null;
  menuNodeId?: string;
  setIsChildOpen: React.Dispatch<React.SetStateAction<boolean>>;
}

const MenuContext = React.createContext<MenuContextProps | null>(null);

export interface MenuProviderProps extends React.HTMLAttributes<HTMLDivElement> {
  trigger?: 'rightClick' | 'click' | 'hover';
  placement?: string;
  isOpen?: boolean;
  defaultIsOpen?: boolean;
  onOpenChange?: (open: boolean) => void;
  showArrow?: boolean;
  children: React.ReactNode;
  atCursor?: boolean;
}

export const MenuComponent = React.forwardRef<HTMLDivElement, MenuProviderProps>((props, ref) => {
  const {
    trigger = 'click',
    placement = '',
    isOpen: controlledIsOpen,
    defaultIsOpen,
    onOpenChange: onOpenChangeProp,
    showArrow = false,
    children,
    atCursor = false,
    ...rest
  } = props;

  const prevContext = React.useContext(MenuContext); // in case there is a parent Menu

  const [isChildOpen, setIsChildOpen] = React.useState(false);
  const [isOpen, setIsOpen] = useControllableState<boolean>({
    prop: controlledIsOpen,
    defaultProp: defaultIsOpen ?? false,
    onChange: onOpenChangeProp,
  }) as [boolean, React.Dispatch<React.SetStateAction<boolean>>];
  const menuIsOpen = isOpen ?? false;

  const [activeIndex, setActiveIndex] = React.useState<number | null>(null);

  const elementsRef = React.useRef<Array<HTMLButtonElement | null>>([]);
  const labelsRef = React.useRef<Array<string | null>>([]);
  const arrowRef = React.useRef(null);

  const nodeId = useFloatingNodeId();
  const parentId = prevContext?.menuNodeId;
  const tree = useFloatingTree();

  const isNested = parentId != null;

  const onOpenChange = (value: boolean, event?: Event, reason?: OpenChangeReason) => {
    if (isChildOpen && !value && reason === 'hover') {
      return;
    }
    setIsOpen(value);
  };

  const middleware = [offset({ mainAxis: isNested ? 0 : 4, alignmentAxis: isNested ? -4 : 0 }), flip(), shift()];

  if (showArrow) {
    middleware.push(arrow({ element: arrowRef }));
  }

  const { x, y, strategy, refs, context } = useFloating<HTMLDivElement>({
    nodeId,
    open: menuIsOpen,
    onOpenChange: onOpenChange,
    placement: placement !== '' ? (placement as Placement) : isNested ? 'right-start' : 'bottom-start',
    middleware,
    whileElementsMounted: autoUpdate,
  });

  const hover = useHover(context, {
    enabled: trigger === 'hover' || isNested,
    delay: { open: 75 },
    handleClose: safePolygon({ requireIntent: false, blockPointerEvents: true }),
  });

  const click = useClick(context, {
    event: 'mousedown',
    enabled: trigger === 'click',
    toggle: true,
  });

  const role = useRole(context, { role: 'menu' });
  const dismiss = useDismiss(context, { bubbles: true });

  const listNavigation = useListNavigation(context, {
    listRef: elementsRef,
    activeIndex,
    onNavigate: setActiveIndex,
    nested: isNested,
    orientation: 'vertical',
  });

  const typeahead = useTypeahead(context, {
    listRef: labelsRef,
    onMatch: menuIsOpen ? setActiveIndex : undefined,
    activeIndex,
  });

  const clientPoint = useClientPoint(context, { enabled: atCursor && !isOpen });

  const interactions = useInteractions([hover, click, role, dismiss, listNavigation, typeahead, clientPoint]);

  const contextValue: MenuContextProps = {
    activeIndex,
    setActiveIndex,
    getItemProps: interactions.getItemProps,
    isOpen: menuIsOpen,
    setIsOpen: onOpenChange,
    trigger,
    floating: { x, y, strategy, refs, context, interactions },
    elementsRef,
    labelsRef,
    isNested,
    parent: prevContext,
    menuNodeId: nodeId,
    setIsChildOpen,
  };

  return <MenuContext.Provider value={contextValue}>{children}</MenuContext.Provider>;
});

export interface MenuProps extends MenuProviderProps {
  label?: string;
  showArrow?: boolean;
}
export const Menu = React.forwardRef<HTMLDivElement, MenuProps>((props, ref) => {
  const parentId = useFloatingParentNodeId();
  const isRoot = parentId === null;
  if (isRoot) {
    return (
      <FloatingTree>
        <MenuComponent {...props} ref={ref}>
          {props.children ? (
            props.children
          ) : props.label ? (
            <>
              <MenuTrigger label={props.label} />
              <MenuContent>{/* Menu-items */}</MenuContent>
            </>
          ) : null}
        </MenuComponent>
      </FloatingTree>
    );
  }
  return (
    <MenuComponent {...props} ref={ref}>
      {props.children ? (
        props.children
      ) : props.label ? (
        <>
          <MenuTrigger label={props.label} />
          <MenuContent>{/* Menu-items */}</MenuContent>
        </>
      ) : null}
    </MenuComponent>
  );
});

export type MenuTriggerProps = Omit<BaseMenuItemProps, 'role' | 'checked' | 'onCheckedChange' | 'closeOnClick'> & {
  x?: number;
  y?: number;
  rightClick?: boolean;
};

export const MenuTrigger = React.forwardRef<HTMLElement, MenuTriggerProps>(
  ({ label, as, disabled, leftSlot, rightSlot, x = null, y = null, children, ...props }, ref) => {
    const context = React.useContext(MenuContext);
    if (!context) throw new Error('MenuTrigger must be used within a Menu');

    const { ref: itemRef, isActive, index } = useMenuItem();
    const isNested = context.isNested;

    if (x != null && y != null) {
      const rect = context.floating.refs.reference.current?.getBoundingClientRect() ?? { x: 0, y: 0 };
      context.floating.x = x + rect.x;
      context.floating.y = y + rect.y;
    }

    const mergedRef = useMergeRefs([context.floating.refs.setReference, ref, itemRef]);
    if (children != null) {
      const referenceProps = {
        ...context.floating.interactions.getReferenceProps({}),
      };

      if (props.rightClick) {
        referenceProps.onContextMenu = (event: React.MouseEvent) => {
          event.preventDefault();
          context.setIsOpen(!context.isOpen);
        };
        delete referenceProps.onMouseDown;
        delete referenceProps.onPointerDown;
        delete referenceProps.onClick;

        if (context.isOpen) {
          referenceProps.onClick = (event: React.MouseEvent) => {
            event.preventDefault();
            context.setIsOpen(false);
          };
        }
      }

      const element = React.cloneElement(children as React.ReactElement<{ ref?: React.Ref<HTMLElement> }>, {
        ...props,
        ...referenceProps,
        ref: mergedRef,
      });

      return element;
    }

    return (
      <BaseMenuItem
        ref={mergedRef}
        as={as ?? 'button'}
        role="menuitem"
        label={label}
        leftSlot={leftSlot}
        rightSlot={
          rightSlot ??
          (isNested ? <Icon size={12} component="icon-[ic--baseline-arrow-forward-ios]" className="text-secondary-500" /> : undefined)
        }
        isTrigger
        disabled={disabled}
        {...props}
        {...context.floating.interactions.getReferenceProps({
          onMouseEnter: () => {
            context.setIsChildOpen(true);
          },
        })}
      />
    );
  },
);

export type MenuContentProps = React.HTMLProps<HTMLDivElement>;
export const MenuContent = React.forwardRef<HTMLDivElement, MenuContentProps>((props, ref) => {
  const context = React.useContext(MenuContext);
  if (!context) throw new Error('MenuContent must be used within a Menu');

  const { floating } = context;
  const { x, y, strategy, interactions } = floating;

  const isNested = context.isNested;
  return (
    <FloatingList elementsRef={context.elementsRef} labelsRef={context.labelsRef}>
      {context.isOpen && (
        <FloatingPortal>
          <FloatingFocusManager context={floating.context} modal={false} initialFocus={isNested ? -1 : 0} returnFocus={!isNested}>
            <div
              ref={node => {
                floating.refs.setFloating(node);
                if (ref) {
                  if (typeof ref === 'function') ref(node);
                  else (ref as React.RefObject<HTMLDivElement | null>).current = node;
                }
              }}
              {...interactions.getFloatingProps(props)}
              style={{ position: strategy, top: y ?? 0, left: x ?? 0 }}
              className={`${styles.menu} ${props.className}`}
            >
              {props.children}
            </div>
          </FloatingFocusManager>
        </FloatingPortal>
      )}
    </FloatingList>
  );
});

interface BaseMenuItemProps extends React.HTMLAttributes<HTMLElement> {
  as?: React.ElementType;
  role?: 'menuitem' | 'menuitemcheckbox' | 'menuitemradio';
  checked?: boolean;
  onCheckedChange?: (checked: boolean) => void;
  label?: string;
  disabled?: boolean;
  leftSlot?: React.ReactNode;
  iconLeft?: string;
  rightSlot?: React.ReactNode;
  iconRight?: string;
  closeOnClick?: boolean;
  isTrigger?: boolean;
}

const BaseMenuItem = React.forwardRef<HTMLElement, BaseMenuItemProps>(
  (
    {
      as: Component = 'button',
      role = 'menuitem',
      checked,
      onCheckedChange,
      label,
      disabled,
      children,
      leftSlot,
      rightSlot,
      iconLeft,
      iconRight,
      closeOnClick = false,
      className,
      isTrigger = false,
      ...props
    },
    forwardedRef,
  ) => {
    const context = React.useContext(MenuContext);
    if (!context) throw new Error('BaseMenuItem must be used within a Menu');

    const { ref, index } = useMenuItem();
    const tree = useFloatingTree();

    const handleClick = (event: React.MouseEvent<HTMLElement>) => {
      event.stopPropagation();
      event.preventDefault();
      props.onClick?.(event);

      if (closeOnClick) {
        tree?.events.emit('click');
        context?.setIsOpen(false);
        context?.parent?.setIsOpen(false);
        context?.parent?.parent?.setIsOpen(false);
        context?.parent?.parent?.parent?.setIsOpen(false);
        context?.parent?.parent?.parent?.parent?.setIsOpen(false);
      }
    };

    const isActive = index === context.activeIndex;

    return (
      <Component
        ref={useMergeRefs([ref, forwardedRef])}
        type={Component === 'button' ? 'button' : undefined}
        role={role as 'menuitem' | 'menuitemcheckbox' | 'menuitemradio'}
        tabIndex={isActive ? 0 : -1}
        data-open={context?.isOpen && isTrigger ? '' : null}
        disabled={Component === 'button' ? disabled : undefined}
        {...props}
        {...context.getItemProps({
          onClick: handleClick,
          onFocus: event => {
            props.onFocus?.(event);
          },
        })}
        className={`${styles.menuItem} ${className ?? ''}`}
      >
        {/* Left Slot */}
        {iconLeft ? <Icon className="text-secondary-500" size={16} component={iconLeft} /> : leftSlot}

        {/* Main Content */}
        <span className={styles.mainSlot}>{label ? label : children}</span>

        {/* Right Slot */}
        {iconRight ? <Icon className="text-secondary-500" size={16} component={iconRight} /> : rightSlot}
      </Component>
    );
  },
);

export interface MenuItemProps extends Omit<BaseMenuItemProps, 'role' | 'onCheckedChange' | 'checked'> {
  closeOnClick?: boolean;
}

export const MenuItem = React.forwardRef<HTMLElement, MenuItemProps>(({ closeOnClick = true, ...props }, ref) => (
  <BaseMenuItem {...props} ref={ref} closeOnClick={closeOnClick} />
));
export interface MenuCheckboxItemProps extends Omit<BaseMenuItemProps, 'role' | 'onCheckedChange' | 'checked'> {
  checked: boolean;
  onCheckedChange: (checked: boolean) => void;
  closeOnClick?: boolean;
}

export const MenuCheckboxItem = React.forwardRef<HTMLButtonElement, MenuCheckboxItemProps>(
  ({ checked, onCheckedChange, leftSlot, closeOnClick = false, ...props }, ref) => (
    <BaseMenuItem
      {...props}
      ref={ref}
      role="menuitemcheckbox"
      leftSlot={
        leftSlot ?? (
          <Icon
            size={16}
            className={checked ? 'text-primary' : 'opacity-50'}
            component={checked ? 'icon-[ic--baseline-check-box]' : 'icon-[ic--baseline-check-box-outline-blank]'}
          />
        )
      }
      onClick={event => {
        event.stopPropagation();
        onCheckedChange(!checked);
        props.onClick?.(event);
      }}
      closeOnClick={closeOnClick}
    />
  ),
);

export interface MenuRadioGroupProps {
  value: string;
  onValueChange: (value: string) => void;
  children: React.ReactNode;
}
const MenuRadioGroupContext = React.createContext<{
  value: string;
  onValueChange: (value: string) => void;
} | null>(null);
export const MenuRadioGroup: React.FC<MenuRadioGroupProps> = ({ value, onValueChange, children }) => {
  return <MenuRadioGroupContext.Provider value={{ value, onValueChange }}>{children}</MenuRadioGroupContext.Provider>;
};

export interface MenuRadioItemProps extends Omit<BaseMenuItemProps, 'role' | 'onCheckedChange' | 'checked'> {
  value: string;
}

export const MenuRadioItem = React.forwardRef<HTMLButtonElement, MenuRadioItemProps>(
  ({ value, leftSlot, closeOnClick = false, ...props }, ref) => {
    const radioContext = React.useContext(MenuRadioGroupContext);
    const selected = radioContext?.value === value;

    return (
      <BaseMenuItem
        {...props}
        ref={ref}
        role="menuitemradio"
        leftSlot={
          leftSlot ?? (
            <Icon
              size={16}
              className={selected ? 'text-primary' : 'opacity-50'}
              component={selected ? 'icon-[ic--baseline-radio-button-checked]' : 'icon-[ic--baseline-radio-button-unchecked]'}
            />
          )
        }
        onClick={event => {
          event.stopPropagation();
          radioContext?.onValueChange(value);
          props.onClick?.(event);
        }}
        closeOnClick={closeOnClick}
      />
    );
  },
);

export function useMenuItem() {
  const context = React.useContext(MenuContext);
  if (!context) throw new Error('useMenuItem must be used within a Menu');

  const item = useListItem();
  return {
    ...item,
    isActive: () => item.index === context.activeIndex,
  };
}

export const MenuSeparator = () => <div className={styles.menuSeparator} />;

export interface MenuLabelProps extends React.HTMLAttributes<HTMLDivElement> {
  label: string;
}

export const MenuLabel: React.FC<MenuLabelProps> = ({ label, className, ...props }) => {
  return (
    <div className={`${styles.menuLabel} ${className || ''}`} {...props}>
      {label}
    </div>
  );
};
