import {
  readInSchemaFromBackend,
  resetGraphQueryResults,
  useActiveQuery,
  useActiveSaveState,
  useAppDispatch,
  useAuthCache,
  useAuthentication,
  useML,
  useQuerybuilderHash,
  useSessionCache,
  useVisualization,
  wsSchemaRequest,
  wsSchemaStatsRequest,
  wsSchemaSubscription,
} from '@/lib/data-access';
import { Broker, wsQueryCountSubscription, wsQueryErrorSubscription, wsQueryRequest, wsQuerySubscription } from '@/lib/data-access/broker';
import { addError, addInfo } from '@/lib/data-access/store/configSlice';
import { setMLResult } from '@/lib/data-access/store/mlSlice';
import { isEqual } from 'lodash-es';
import { useEffect } from 'react';
import {
  AbortedError,
  GraphQueryTranslationResultMessage,
  LinkPredictionInstance,
  SaveState,
  Schema,
  allMLTypes,
  wsReturnKey,
} from 'ts-common';
import {
  DatabaseStatus,
  nilUUID,
  wsGetState,
  wsGetStateSubscription,
  wsGetStates,
  wsGetStatesSubscription,
  wsTestSaveStateConnection,
  wsTestSaveStateConnectionSubscription,
  wsUpdateState,
} from '../broker/wsState';
import { wsReconnectSubscription, wsUserGetPolicy } from '../broker/wsUser';
import { authorized, setSessionID } from '../store/authSlice';
import { abortedBackend, asyncSetNewGraphQueryResult, queryingBackend, setNewGraphQueryCountResult } from '../store/graphQuerySlice';
import { unSelect } from '../store/interactionSlice';
import { clearSchema, setSchemaAttributeInformation, setSchemaLoading, setSchemaState } from '../store/schemaSlice';
import {
  addSaveState,
  selectSaveState,
  setActiveQueryID,
  setFetchingSaveStates,
  setQueryState,
  setQuerybuilderNodes,
  testedSaveState,
  updateSaveStateList,
  updateSelectedSaveState,
} from '../store/sessionSlice';
import { VisState, setVisualizationState } from '../store/visualizationSlice';
import { URLParams, getParam } from './url';

export const EventBus = (props: { onRunQuery: (useCached: boolean) => void; onAuthorized: () => void }) => {
  const { login } = useAuthentication();
  const auth = useAuthCache();
  const dispatch = useAppDispatch();
  const ml = useML();
  const session = useSessionCache();
  const queryHash = useQuerybuilderHash();
  const activeSS = useActiveSaveState();
  const activeQuery = useActiveQuery();
  const visState = useVisualization();

  async function loadSaveState(saveStateID: string | undefined, saveStates: Record<string, SaveState>) {
    if (saveStateID && saveStates && saveStateID in saveStates) {
      console.debug('Setting state fetched from database', saveStateID, saveStates);
      const state = saveStates[saveStateID];
      if (state) {
        if (state.queryStates && state.queryStates.openQueryArray.length > 0) {
          let activeQuery = state.queryStates.openQueryArray.find(q => q.id === state.queryStates.activeQueryId);
          if (!activeQuery) {
            activeQuery = state.queryStates.openQueryArray[0];
            dispatch(setActiveQueryID(activeQuery.id || -1));
          }
          if (activeQuery) {
            const queryBuilderState = {
              ...state.queryStates,
              openQueryArray: state.queryStates.openQueryArray.map(query => ({
                ...query,
                ignoreReactivity: false,
                queryTranslation: {} as GraphQueryTranslationResultMessage,
              })),
            };

            dispatch(setQueryState(queryBuilderState));
            dispatch(setQuerybuilderNodes(activeQuery));
          }
        } else {
          console.error('No queries found in state', state);
        }
        dispatch(
          setVisualizationState(
            Object.keys(state.visualizations).length !== 0 // should only occur in mock data
              ? (state.visualizations as VisState)
              : {
                  activeVisualizationIndex: -1,
                  openVisualizationArray: [],
                },
          ),
        );

        if (state.schemas.length > 0) {
          dispatch(setSchemaState(state.schemas[0] as Schema));
        }

        testConnection(state.id);
      }
    }
  }

  function testConnection(saveStateID: string) {
    dispatch(testedSaveState({ saveStateID, status: DatabaseStatus.testing }));
    wsTestSaveStateConnection({ saveStateID });
  }

  useEffect(() => {
    const unsubs: (() => void)[] = [];

    unsubs.push(
      wsSchemaSubscription(({ data, status }) => {
        if (status !== 'success') {
          dispatch(addError('Failed to fetch schema'));
          dispatch(clearSchema({ error: true }));
          return;
        }
        if (!data) {
          dispatch(addError('Failed to fetch schema'));
          return;
        } else {
          dispatch(readInSchemaFromBackend(data));
        }
      }),
    );

    unsubs.push(
      wsQuerySubscription(({ data, status, callID }) => {
        if (status === 'aborted') {
          dispatch(abortedBackend({ type: 'result', callID: callID }));
          return;
        }
        if (!data || status !== 'success') {
          dispatch(addError('Failed to fetch graph query result'));
          return;
        }
        asyncSetNewGraphQueryResult({ data, callID }, dispatch);
        dispatch(unSelect());
      }),
    );

    unsubs.push(
      wsQueryCountSubscription(params => {
        const { data, status } = params;
        if (status === 'aborted') {
          dispatch(abortedBackend({ type: 'count', callID: params.callID }));
          return;
        }
        if (!data || status !== 'success') {
          console.warn('Failed to fetch graph query counts', params);
          dispatch(addError('Failed to fetch graph query counts'));
          return;
        }
        dispatch(setNewGraphQueryCountResult({ data, callID: params.callID }));
      }),
    );

    unsubs.push(
      wsQueryErrorSubscription(params => {
        if (params.data?.name === new AbortedError().name) {
          dispatch(resetGraphQueryResults());
          return;
        }
        dispatch(addError('Failed to fetch graph query result'));
        console.error('Query Error', params.data);
        dispatch(resetGraphQueryResults());
      }),
    );

    unsubs.push(
      wsTestSaveStateConnectionSubscription(params => {
        const data = params.data;
        if (data?.status === 'success' && data.saveStateID) {
          dispatch(
            testedSaveState({
              saveStateID: data.saveStateID,
              status: data.status === 'success' ? DatabaseStatus.online : DatabaseStatus.offline,
            }),
          );
        } else {
          dispatch(addError('DB connection failed'));
        }
      }),
    );

    unsubs.push(
      wsReconnectSubscription(params => {
        const data = params.data;
        if (data == null) return;

        if (!Broker.instance().checkSessionID(data.sessionID)) {
          dispatch(addError('DEBUG: session id does not match!!'));
          throw new Error('Session ID does not match!!');
        }

        dispatch(setSessionID(data));

        props.onAuthorized();

        dispatch(setFetchingSaveStates(true));

        // check authorizations
        wsUserGetPolicy(null, params => {
          if (params.data) dispatch(authorized(params.data));
        });

        wsGetStates(null, params => {
          const states = params.data;
          if (states) {
            dispatch(setFetchingSaveStates(false));
            params.defaultCallback();

            // Process URL Params after states are fetched
            const paramSaveState = getParam(URLParams.saveState);
            if (paramSaveState && paramSaveState !== nilUUID) {
              // check if the shared savestate is already in the users states
              const stateExists = states.some(state => state.id === paramSaveState);
              if (!stateExists) {
                console.log('Shared save state not found in user states. Fetching:', paramSaveState);
                wsGetState({ saveStateID: paramSaveState }, ({ data, status }) => {
                  if (!['success', 'unchanged'].includes(status) || !data) {
                    dispatch(addError('Failed to fetch state'));
                    return;
                  }
                  if (data.id !== nilUUID) {
                    if (isEqual(data, session.saveStates[data.id])) {
                      console.log('State unchanged, not updating');
                      return;
                    }
                    dispatch(addSaveState({ ss: data, select: true }));
                  }
                });
              } else {
                dispatch(selectSaveState(paramSaveState));
              }
            }
          } else {
            dispatch(addError('Failed to fetch states'));
          }
        });

        // Broker.instance().sendMessage({ //TODO!!
        //   sessionID: auth?.sessionID || '',
        //   key: 'broadcastState',
        //   body: { type: 'subscribe', status: '', value: {} },
        // });
      }),
    );

    Broker.instance().subscribe(params => {
      if (!params.data) {
        dispatch(addError('Failed to fetch schema stats'));
        return;
      }
      dispatch(setSchemaAttributeInformation(params.data));
      dispatch(addInfo('Received attribute information'));
    }, wsReturnKey.schemaStatsResult);

    // Broker.instance().subscribe((data: QueryBuilderState) => dispatch(setQuerybuilderNodes(data)), 'query_builder_state');

    allMLTypes.forEach(mlType => {
      const id = Broker.instance().subscribe(params => {
        const data = params.data as unknown as LinkPredictionInstance[];
        dispatch(setMLResult({ type: mlType, result: data }));
      }, mlType);
      unsubs.push(() => Broker.instance().unSubscribe(mlType, id));
    });

    unsubs.push(
      wsGetStatesSubscription(params => {
        console.debug('Save States updated', params.data, params.status);
        if (params.status !== 'success' || !params.data) {
          dispatch(addError('Failed to fetch states'));
          return;
        }

        dispatch(updateSaveStateList(params.data));
      }),
    );

    unsubs.push(
      wsGetStateSubscription(({ data, status }) => {
        if (!['success', 'unchanged'].includes(status) || !data) {
          dispatch(addError('Failed to fetch state'));
          return;
        }
        if (data.id !== nilUUID) {
          if (isEqual(data, session.saveStates[data.id])) {
            console.log('State unchanged, not updating');
            return;
          }
          dispatch(addSaveState({ ss: data, select: false }));
        }
      }),
    );

    login();

    // Setup cleanup
    return () => {
      // clear callback subscriptions
      unsubs.forEach(unsub => {
        unsub();
      });
      Broker.instance().unSubscribeAll(wsReturnKey.queryStatusTranslationResult);
      Broker.instance().unSubscribeAll(wsReturnKey.queryStatusResult);
      Broker.instance().unSubscribeAll(wsReturnKey.schemaInference);
    };
  }, []);

  useEffect(() => {
    if (activeSS && activeSS.id in session.testedSaveState) {
      console.log('Query State Hash Updated', queryHash, activeSS.id);
      wsUpdateState(activeSS, ({ data }) => {
        // if (status === 'unchanged') {
        //  TODO: check if this is needed, but for now removed so that the query is always sent, solving issue with changing ss
        //   console.log('State unchanged, not updating');
        //   return;
        // }

        if (data) {
          if (activeQuery?.id) {
            const callID = Broker.instance().generateCallID();
            dispatch(queryingBackend({ callID, status: 'querying' }));
            wsQueryRequest({ saveStateID: activeSS.id, ml, queryID: activeQuery.id, useCached: true, callID });
          }
        }
      });
      // }
    }
  }, [queryHash]);

  useEffect(() => {
    if (session.currentSaveState) {
      const state = { ...session.saveStates[session.currentSaveState] };
      if (!isEqual(state.visualizations, visState)) {
        console.debug('Updating visState state', visState);
        state.visualizations = { ...visState };
        dispatch(updateSelectedSaveState(state));
        wsUpdateState(state);
      }
    }
  }, [visState]);

  useEffect(() => {
    // New active database
    if (session.currentSaveState && session.currentSaveState !== nilUUID) {
      dispatch(setSchemaLoading(true));
      wsSchemaRequest({ saveStateID: session.currentSaveState, useCached: true });
      wsSchemaStatsRequest({ saveStateID: session.currentSaveState });
      loadSaveState(session.currentSaveState, session.saveStates);
    }
  }, [session.currentSaveState]);

  useEffect(() => {
    // New active database
    if (
      session.currentSaveState &&
      session.testedSaveState[session.currentSaveState] === DatabaseStatus.online &&
      activeQuery?.id != null
    ) {
      const callID = Broker.instance().generateCallID();
      dispatch(queryingBackend({ callID, status: 'querying' }));
      wsQueryRequest({ saveStateID: session.currentSaveState, ml, queryID: activeQuery.id, useCached: true, callID });
    }
  }, [session.currentSaveState, session.testedSaveState]);

  useEffect(() => {
    // Newly (un)authorized
    if (auth.authentication?.authenticated && auth.authentication?.jwt) {
      console.warn('Authorized', auth.authentication);
      if (Broker.instance().isConnected()) {
        console.log('WS already connected, no need to reconnect');
        props.onAuthorized();
      } else {
        console.log('WS not connected, connecting');
        Broker.instance()
          .useAuth(auth.authentication)
          .connect(() => {
            console.debug('WS connected', session.currentSaveState, window.location.search);
          });
      }
    } else {
      // dispatch(logout());
    }
  }, [auth.authentication]);

  return <div className="hide"></div>;
};
