import { QueryMultiGraphology } from '@/lib/querybuilder';
import { createSlice, PayloadAction } from '@reduxjs/toolkit';
import { MultiGraph } from 'graphology';
import { isEqual } from 'lodash-es';
import objectHash from 'object-hash';
import {
  defaultGraph,
  Query,
  QueryBuilderSettings,
  QueryGraphEdgeHandle,
  QueryMultiGraph,
  QueryState,
  QueryUnionType,
  SaveStateAuthorizationHeaders,
  SaveStateWithAuthorization,
} from 'ts-common';
import { getParam, setParam, URLParams } from '../api/url';
import { wsUpdateQueryOrder } from '../broker/wsQuery';
import { DatabaseStatus } from '../broker/wsState';
import type { RootState } from './store';

/** Message format of the error message from the backend */
export type ErrorMessage = {
  errorStatus: any;
  queryID: any;
};

/** Cache type */
export type SessionCacheI = {
  currentSaveState?: string; // id of the current save state
  saveStates: Record<string, SaveStateWithAuthorization>;
  fetchingSaveStates: boolean;
  testedSaveState: Record<string, DatabaseStatus>;
};

const newStateAuthorizationHeaders: SaveStateAuthorizationHeaders = {
  query: { W: true, R: true },
  database: { W: true, R: true },
  visualization: { W: true, R: true },
  schema: { W: true, R: true },
  savestate: { W: true, R: true },
};

// Define the initial state using that type
export const initialState: SessionCacheI = {
  currentSaveState: undefined,
  saveStates: {},
  fetchingSaveStates: true, // default to true to prevent flashing of the UI
  testedSaveState: {},
};

export const sessionSlice = createSlice({
  name: 'session',
  initialState: initialState,
  reducers: {
    setFetchingSaveStates: (state: SessionCacheI, action: PayloadAction<boolean>) => {
      state.fetchingSaveStates = action.payload;
    },
    selectSaveState: (state: SessionCacheI, action: PayloadAction<string | undefined>) => {
      if (action.payload && action.payload in state.saveStates) {
        state.currentSaveState = action.payload;
      } else {
        console.error('Save state not found');
        state.currentSaveState = undefined;
      }
      setParam(URLParams.saveState, action.payload);
    },
    updateSelectedSaveState: (state: SessionCacheI, action: PayloadAction<SaveStateWithAuthorization>) => {
      if (state.currentSaveState === action.payload.id && state.currentSaveState) state.saveStates[state.currentSaveState] = action.payload;
    },
    updateSaveStateList: (state: SessionCacheI, action: PayloadAction<SaveStateWithAuthorization[]>) => {
      // Does NOT clear the states, just adds in new data
      const newState: Record<string, SaveStateWithAuthorization> = {};
      if (!action.payload) return;
      action.payload.forEach(ss => {
        newState[ss.id] = ss;
      });
      state.saveStates = { ...state.saveStates, ...newState };
    },
    setSaveStateList: (state: SessionCacheI, action: PayloadAction<SaveStateWithAuthorization[]>) => {
      // Clears the states and puts in new data
      const newState: Record<string, SaveStateWithAuthorization> = {};
      action.payload.forEach(ss => {
        newState[ss.id] = ss;
      });
      state.saveStates = newState;
      const paramSaveState = getParam(URLParams.saveState);
      if (!state.currentSaveState) {
        if (paramSaveState && paramSaveState in state.saveStates) {
          state.currentSaveState = paramSaveState;
          setParam(URLParams.saveState, paramSaveState);
        } else if (Object.keys(state.saveStates).length > 0) {
          state.currentSaveState = Object.keys(state.saveStates)[0];
          setParam(URLParams.saveState, state.currentSaveState);
        } else {
          state.currentSaveState = undefined;
          setParam(URLParams.saveState, undefined);
        }
      }
    },
    addSaveState: (state: SessionCacheI, action: PayloadAction<{ ss: SaveStateWithAuthorization; select: boolean }>) => {
      if (state.saveStates === undefined) state.saveStates = {};
      state.saveStates[action.payload.ss.id] = action.payload.ss;
      if (action.payload.select === true) {
        state.currentSaveState = action.payload.ss.id;
        setParam(URLParams.saveState, action.payload.ss.id);
      }
    },
    deleteSaveState: (state: SessionCacheI, action: PayloadAction<string>) => {
      delete state.saveStates[action.payload];
      if (state.currentSaveState === action.payload) {
        if (Object.keys(state.saveStates).length > 0) {
          state.currentSaveState = Object.keys(state.saveStates)[0];
          setParam(URLParams.saveState, state.currentSaveState);
        } else {
          state.currentSaveState = undefined;
          setParam(URLParams.saveState, undefined);
        }
      }
    },
    testedSaveState: (state: SessionCacheI, action: PayloadAction<{ saveStateID: string; status: DatabaseStatus }>) => {
      state.testedSaveState = { ...state.testedSaveState, [action.payload.saveStateID]: action.payload.status };
    },
    setQuerybuilderGraph: (state: SessionCacheI, action: PayloadAction<QueryMultiGraph>) => {
      const activeQuery = getActiveQuery(state);
      if (activeQuery) {
        activeQuery.graph = {
          ...action.payload,
        };
        activeQuery.graphCounts = { nodeCounts: { updatedAt: 0 } };
      }
    },
    setQuerybuilderSettings: (state: SessionCacheI, action: PayloadAction<QueryBuilderSettings>) => {
      const activeQuery = getActiveQuery(state);
      if (activeQuery) {
        activeQuery.settings = action.payload;
      }
    },
    attributeShownToggle: (state: SessionCacheI, action: PayloadAction<QueryGraphEdgeHandle>) => {
      const activeQuery = getActiveQuery(state);
      if (activeQuery) {
        const existing = activeQuery.attributesBeingShown.findIndex(a => isEqual(a, action.payload));
        if (existing === -1) {
          activeQuery.attributesBeingShown.push(action.payload);
        } else {
          activeQuery.attributesBeingShown.splice(existing, 1);
        }
      }
    },
    setQueryUnionType: (state: SessionCacheI, action: PayloadAction<{ nodeId: string; unionType: QueryUnionType }>) => {
      const activeQuery = getActiveQuery(state);
      if (activeQuery) {
        if (activeQuery.settings.unionTypes == null) {
          activeQuery.settings.unionTypes = {};
        }
        activeQuery.settings.unionTypes[action.payload.nodeId] = action.payload.unionType;
      }
    },
    setQueryName: (state: SessionCacheI, action: PayloadAction<string>) => {
      const activeQuery = getActiveQuery(state);
      if (activeQuery) {
        activeQuery.name = action.payload;
      }
    },
    addNewQuery: (state: SessionCacheI, payload: PayloadAction<Query>) => {
      if (!payload.payload.id) {
        console.error('No query id found');
        return;
      }

      const activeSS = getActiveSaveState(state);
      if (!activeSS) {
        console.error('No active save state found');
        return;
      }

      activeSS.queryStates.openQueryArray.push(payload.payload);
      activeSS.queryStates.activeQueryId = payload.payload.id;
    },
    removeQueryByID: (state: SessionCacheI, action: PayloadAction<number>) => {
      const activeSS = getActiveSaveState(state);
      if (!activeSS) {
        console.error('No active save state found');
        return;
      }
      activeSS.queryStates.openQueryArray = activeSS.queryStates.openQueryArray.filter(query => query.id !== action.payload);
      if (activeSS.queryStates.activeQueryId === action.payload) {
        // Set the active query to the first query if the active query was removed
        const availableQuery = activeSS.queryStates.openQueryArray.find(query => query.id !== action.payload);
        activeSS.queryStates.activeQueryId =
          availableQuery && availableQuery.id ? availableQuery.id : activeSS.queryStates.openQueryArray[0]?.id || -1;
        if (activeSS.queryStates.activeQueryId === -1) {
          console.error('No active query found 1');
        }
      }
    },
    setActiveQueryID: (state: SessionCacheI, action: PayloadAction<number>) => {
      const activeSS = getActiveSaveState(state);
      if (!activeSS) {
        console.error('No active save state found');
        return;
      }
      if (activeSS.queryStates.openQueryArray.find(query => query.id === action.payload) === undefined) {
        const availableQuery = activeSS.queryStates.openQueryArray.find(query => query.id !== action.payload);
        activeSS.queryStates.activeQueryId = availableQuery && availableQuery.id ? availableQuery.id : -1;
        console.error('QueryID not found', action.payload, activeSS.queryStates.openQueryArray);
        return;
      }
      activeSS.queryStates.activeQueryId = action.payload;
    },
    setQueryState: (state: SessionCacheI, action: PayloadAction<QueryState>) => {
      const activeSS = getActiveSaveState(state);
      if (!activeSS) {
        console.error('No active save state found');
        return;
      }

      activeSS.queryStates = action.payload;
    },
    reorderQueryState: (
      state,
      action: PayloadAction<{
        oldIndex: number;
        newIndex: number;
      }>,
    ) => {
      const activeSS = getActiveSaveState(state);
      const queryState = activeSS?.queryStates;

      if (activeSS == null || queryState == null) {
        console.error('No query state found');
        return;
      }

      const { oldIndex, newIndex } = action.payload;

      if (oldIndex >= 0 && oldIndex < queryState.openQueryArray.length && newIndex >= 0 && newIndex < queryState.openQueryArray.length) {
        const [movedItem] = queryState.openQueryArray.splice(oldIndex, 1);
        queryState.openQueryArray.splice(newIndex, 0, movedItem);

        wsUpdateQueryOrder({
          saveStateID: state.currentSaveState!,
          order: Object.fromEntries(queryState.openQueryArray.map((query, i) => [query.id, i])),
        });
      }
    },
    /**
     * Sets the querybuilder nodes, settings, and attributes being shown,
     * if the payload contains the required information.
     * @param {QueryBuilderState} action.payload the payload with the new state
     */
    setQuerybuilderNodes: (state: SessionCacheI, action: PayloadAction<Query>) => {
      const activeQuery = getActiveQuery(state);

      if (activeQuery && action.payload.graph?.nodes && action.payload.graph?.edges) {
        activeQuery.graph = action.payload.graph;
        activeQuery.settings = action.payload.settings;
        activeQuery.attributesBeingShown = action.payload.attributesBeingShown || [];
      }
    },
    clearQB: (state: SessionCacheI) => {
      const activeQuery = getActiveQuery(state);
      if (activeQuery) {
        activeQuery.graph = defaultGraph();
        activeQuery.settings = {
          limit: 500,
          depth: { min: 1, max: 1 },
          layout: 'manual',
          autocompleteRelation: true,
          unionTypes: {},
        };
        activeQuery.attributesBeingShown = [];
      } else {
        console.error('No active query found 2');
      }
    },
  },
});

export const {
  selectSaveState,
  deleteSaveState,
  updateSaveStateList,
  setSaveStateList,
  addSaveState,
  testedSaveState,
  setFetchingSaveStates,
  updateSelectedSaveState,
  setQuerybuilderGraph,
  setQuerybuilderSettings,
  setActiveQueryID,
  addNewQuery,
  setQueryName,
  clearQB,
  removeQueryByID,
  setQueryState,
  reorderQueryState,
  setQuerybuilderNodes,
  attributeShownToggle,
  setQueryUnionType,
} = sessionSlice.actions;

function getActiveSaveState(sessionCache: SessionCacheI): SaveStateWithAuthorization | undefined {
  return sessionCache.saveStates?.[sessionCache.currentSaveState!];
}

function getActiveQuery(state: SessionCacheI): Query | undefined {
  const ss = getActiveSaveState(state);
  if (!ss || !ss.queryStates || !ss.queryStates.activeQueryId || ss.queryStates.activeQueryId === -1) {
    return; // Return an empty object if no active query exists
  }
  const activeQuery = ss.queryStates.openQueryArray.find(q => q.id === ss.queryStates.activeQueryId);
  if (!activeQuery) {
    console.error('No active query found, returning first query');
    if (ss.queryStates.openQueryArray.length === 0) {
      console.error('No queries found');
      return;
    }
    if (ss.queryStates.openQueryArray[0]?.id == null || ss.queryStates.openQueryArray[0]?.id === -1) {
      console.error('First query has no id');
      return;
    }
    sessionSlice.actions.setActiveQueryID(ss.queryStates.openQueryArray[0].id);
    return ss.queryStates.openQueryArray[0];
  }
  return activeQuery;
}

/** Select the querybuilder nodes in serialized fromat */
export const toQuerybuilderGraphology = (graph: QueryMultiGraph): QueryMultiGraphology => {
  const ret = new QueryMultiGraphology();
  ret.import(MultiGraph.from(graph).export());
  return ret;
};

// Other code such as selectors can use the imported `RootState` type
export const sessionCacheState = (state: RootState) => state.sessionCache;
export const activeSaveState = (state: RootState): SaveStateWithAuthorization | undefined => getActiveSaveState(state.sessionCache);

export default sessionSlice.reducer;

export const activeSaveStateQuery = (state: RootState): Query | undefined => getActiveQuery(state.sessionCache);

export const setQuerybuilderGraphology = (payload: QueryMultiGraphology) => {
  return sessionSlice.actions.setQuerybuilderGraph(payload.export());
};

/** Select the querybuilder nodes and convert it to a graphology object */
export const selectQuerybuilderHash = (state: RootState): string => {
  const activeQuery = activeSaveStateQuery(state);

  if (!activeQuery?.graph) {
    return ''; // Return an empty string if no active graph exists
  }

  function isNotEmptyLogicPill(node: any) {
    if (node == null) {
      return false; // is empty
    } else if (node.attributes.type !== 'logic') {
      return true; // is not a logic
    } else if (node.attributes.logic.numExtraInputs === 0) {
      return true; // is a logic with no extra inputs required
    } else if (node.attributes.inputs == null || Object.values(node.attributes.inputs).some(i => i == null || i === '')) {
      return false; // there is a missing input to be filled in, so assume empty
    }

    return true; // default
  }

  const hashedNodes = activeQuery.graph.nodes.filter(isNotEmptyLogicPill).map(n => {
    const node = { ...n };
    if (n?.attributes) {
      const newAttributes = { ...n?.attributes };
      newAttributes.x = 0;
      newAttributes.y = 0;
      newAttributes.height = 0;
      newAttributes.width = 0;
      node.attributes = newAttributes;
    }
    return node;
  });

  const ret = {
    nodes: hashedNodes,
    edges: activeQuery.graph.edges
      .filter(e => {
        const targetNode = activeQuery.graph.nodes.find(n => n.key == e.target);
        return isNotEmptyLogicPill(targetNode);
      })
      .map(n => ({
        key: n.key,
        source: n.source,
        target: n.target,
      })),
    settings: activeQuery.settings,
  };
  return objectHash(ret);
};
