import { wsAddQuery, wsUpdateQuery } from '@/lib/data-access/broker';
import { addError } from '@/lib/data-access/store/configSlice';
import { useState } from 'react';
import { PartialSaveState, SaveState, SaveStateWithAuthorization } from 'ts-common';
import {
  DatabaseStatus,
  useAppDispatch,
  useAuthCache,
  useProjects,
  wsCreateState,
  wsTestDatabaseConnection,
  wsUpdateState,
} from '../../data-access';
import { setSchemaLoading } from '../../data-access/store/schemaSlice';
import { addSaveState, selectSaveState, testedSaveState } from '../../data-access/store/sessionSlice';

export type ConnectionStatus = {
  updating: boolean;
  status: null | string;
  verified: boolean | null;
};

export const useHandleDatabase = () => {
  const dispatch = useAppDispatch();
  const [connectionStatus, setConnectionStatus] = useState<ConnectionStatus>({
    updating: false,
    status: null,
    verified: null,
  });
  const authCache = useAuthCache();
  const project = useProjects();

  async function submitDatabaseChange(
    saveStateData: PartialSaveState,
    type: 'add' | 'update',
    forceAdd: boolean = false,
    concludedCallback: () => void,
  ): Promise<void> {
    setConnectionStatus(() => ({
      updating: true,
      status: 'Testing database connection',
      verified: null,
    }));
    const newSaveStateData = JSON.parse(JSON.stringify(saveStateData)) as PartialSaveState; // undo readonly

    const currentProject = project.currentProject;
    if (currentProject) {
      newSaveStateData.project = currentProject;
    } else {
      newSaveStateData.project = null;
    }

    wsTestDatabaseConnection(newSaveStateData.dbConnections?.[0], params => {
      const data = params.data;
      if (!newSaveStateData) {
        console.error('formData is null');
        return;
      }
      if (newSaveStateData.userId !== authCache.authentication?.userID && authCache.authentication?.userID) {
        console.error('user_id is not equal to auth.userID');
        newSaveStateData.userId = authCache.authentication.userID;
      }
      if (data && data.status === 'success') {
        setConnectionStatus(prevState => ({
          updating: false,
          status: 'Database connection verified',
          verified: true,
        }));
        dispatch(setSchemaLoading(true));
        dispatch(selectSaveState(undefined));
        if (type === 'add' || forceAdd) {
          // Reset query state, backend should handle creating this anew
          const queryStateToClone = JSON.parse(JSON.stringify(newSaveStateData.queryStates ?? {}));
          delete newSaveStateData.queryStates;

          wsCreateState(newSaveStateData, params => {
            (async (newSaveState: SaveStateWithAuthorization) => {
              if (!newSaveState) {
                dispatch(addError('Failed to create new save state.'));
                return;
              }

              // Only if we are adding a save state that is cloned and contains a queryState
              if (newSaveStateData.queryStates != null) {
                // Clone query state
                for (const i in queryStateToClone.openQueryArray) {
                  const query = queryStateToClone.openQueryArray[i];
                  delete query.id;

                  await new Promise<void>((resolve, reject) => {
                    if (Number(i) == 0) {
                      query.id = newSaveState.queryStates.openQueryArray[0].id;
                      queryStateToClone.activeQuery = query.id;
                      wsUpdateQuery({ saveStateID: newSaveState.id, query: query }, params => {
                        if (params.data == null) return reject();
                        return resolve();
                      });
                    } else {
                      wsAddQuery({ saveStateID: newSaveState.id }, params => {
                        const newQuery = params.data;
                        if (newQuery == null) {
                          return reject('Failed to create new save state (query state).');
                        }
                        query.id = newQuery.id;
                        queryStateToClone.activeQueryId = query.id;

                        wsUpdateQuery({ saveStateID: newSaveState.id, query: query }, params => {
                          if (params.data == null) reject();
                          return resolve();
                        });
                      });
                    }
                  });
                }

                newSaveState.queryStates = queryStateToClone;
              }

              dispatch(addSaveState({ ss: newSaveState, select: true }));
              setConnectionStatus({
                updating: false,
                status: null,
                verified: null,
              });
              concludedCallback();
            })(params.data);
          });
        } else {
          if (data.saveStateID) {
            dispatch(testedSaveState({ saveStateID: data.saveStateID, status: DatabaseStatus.untested }));
          }
          wsUpdateState(newSaveStateData as SaveState, params => {
            const updatedSaveState = params.data;

            if (!updatedSaveState) {
              dispatch(addError('Failed to update save state.'));
              return;
            }
            dispatch(addSaveState({ ss: updatedSaveState, select: true }));
            setConnectionStatus({
              updating: false,
              status: null,
              verified: null,
            });
            concludedCallback();
          });
        }
      } else {
        setConnectionStatus(prevState => ({
          updating: false,
          status: 'Database connection test failed',
          verified: false,
        }));
      }
    });
  }
  return { submitDatabaseChange, connectionStatus };
};
